/************************************************************************
 *   MODULE  : MAKE_RDL.C
 *
 *   PURPOSE : builds the radials along the LOS mask
 *
 *   AUTHOR  : Ron Hawkins & Craig Woodfolk, SAIC (original)
 *             Kevin Russo, SAIC (1, 3)
 *             Melissa Mathews, SAIC (2)
 *
 *   Change     Date          Reason
 *   ------     ----          ------
 *   Original   25 Sep 92     n/a
 *   1          20 Oct 92     overall program update
 *   2          17 Jun 93     converted xvt 2.12 calls to xvt 3.01
 *   3          21 Jul 93     implemented new TEC LOS algorithm
 *
 ************************************************************************/

/****************
 * header files *
 ****************/

#include "los.h"

/**********************************
 * file scope function prototypes *
 **********************************/

#ifdef PROTO
static void make_one_radial (GFLOAT radial_azimuth, GFLOAT quad1,
                                GFLOAT quad2, GFLOAT quad3, GFLOAT quad4);
#else
static void make_one_radial ();
#endif

/********************
 * global variables *
 ********************/

extern PARAM_CALC     param;             /* calculations location       */
extern SETTINGS_TYPE  settings;          /* user input parameters       */
extern HANDLES        handle;            /* ghandle array               */
extern RASTER         *prod_raster;      /* 1st product pointer         */
extern RASTER         elev_raster;       /* dted elevation raster       */
extern TRIG_TYPE      trig_table;        /* sine and cosine table hdl   */
extern WIN_INFO       wininfo;           /* window information          */

char  pHMEM prod1_ptr;                   /* product raster data         */

BYTE pHMEM num_hidden;
BYTE pHMEM num_visible;
BYTE pHMEM num_prhidden;
BYTE pHMEM num_prvisible;

/************************
 * file scope variables *
 ************************/

static float pHMEM distance;             /* radial distances array      */
static short pHMEM elev_hgts;            /* radial heights array        */
static long  pHMEM prod_buff_xdim;       /* elevation buffer x-dimen    */
static long  pHMEM prod_buff_ydim;       /* elevation buffer y-dimen    */
static char  pHMEM profile_color;        /* profile color array         */
static short pHMEM dted;                 /* ptr to dted data            */
static GFLOAT *sine, *cosine;            /* ptrs to sine/cosine tables  */

/************************************************************************
 *   FUNCTION     : make_radials
 *
 *   PURPOSE      : constructs the profile or line of sight product
 *
 *   INPUTS       : none
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : do_prod in module DO_PROD.C
 *
 *   CALLS        : mem_alloc in module MEM_ALLC.C
 *                  make_one_radial in module MAKE_RDL.C
 *
 ************************************************************************/

#ifdef PROTO
void make_radials(void)
#else
void make_radials()
#endif
{
   GFLOAT  azimuth_start;                /* beginning azimuth (dec degr)*/
   GFLOAT  azimuth_end;                  /* ending azimuth (dec degr)   */
   GFLOAT  radial_azimuth;               /* current radial azimuth      */
   GFLOAT  dif_lat, dif_lon;             /* latitude & longitude delta  */
   GFLOAT  dif_lon_m;                    /* longitude diff in meters    */
   GFLOAT  dif_lat_m;                    /* latitude diff in meters     */
   GFLOAT  lat, lon;                     /* observer lat and lon        */
   GFLOAT  percent;                      /* percent complete            */
   GFLOAT  quad1;                        /* 1st quadrant angle          */
   GFLOAT  quad2;                        /* 2nd quadrant angle          */
   GFLOAT  quad3;                        /* 3rd quadrant angle          */
   GFLOAT  quad4;                        /* 4th quadrant angle          */

   BYTE    maxcheck;

   GHANDLE hidden, visible, prhidden, prvisible;

   long    loop;                         /* loop counter                */
   long    next_step;                    /* percent done                */
   char    message[80];                  /* status message              */

   long size;

   /*   extern WINDOW los_win;*/

   /* allocate memory for various arrays */
   mem_alloc((long) (param.max_nhts_in_radial + ONE) * sizeof(float),
             &handle.dist, (void *pHMEM) &distance);
   mem_alloc((long) (param.max_nhts_in_radial + ONE) * sizeof(short),
             &handle.hgts, (void *pHMEM) &elev_hgts);
   mem_alloc((long) prod_raster->data_matrix.width * sizeof(long),
             &handle.xdim, (void *pHMEM) &prod_buff_xdim);
   mem_alloc((long) prod_raster->data_matrix.height * sizeof(long),
             &handle.ydim, (void *pHMEM) &prod_buff_ydim);

   /* lock down trig tables */
   sine   = (GFLOAT *)xvt_gmem_lock(trig_table.shandle);
   cosine = (GFLOAT *)xvt_gmem_lock(trig_table.chandle);
   if ((sine == (GFLOAT *)NULL) || (cosine == (GFLOAT *)NULL))
      xvt_dm_post_fatal_exit("Unable to access trig table memory");

   /* lock the dted memory */
   dted = (short pHMEM)xvt_gmem_lock(elev_raster.data_matrix.ghandle);
   if ( dted == (short pHMEM) NULL )
      xvt_dm_post_fatal_exit("unable to lock elev_raster in make_radials()");

   /* lock the product memory */
   prod1_ptr = (char pHMEM) xvt_gmem_lock (prod_raster->data_matrix.ghandle);
   if (prod1_ptr == (char pHMEM)NULL)
      xvt_dm_post_fatal_exit("Unable to lock down product memory in make_radials()");

   /* build the product check arrays */
   size = prod_raster->data_matrix.width *
          prod_raster->data_matrix.height * sizeof(BYTE);

   mem_alloc(size, &visible, (void *pHMEM) &num_visible);

   mem_alloc(size, &hidden, (void *pHMEM) &num_hidden);

   if (settings.prod_type == VIP)
   {
      mem_alloc(size, &prhidden, (void *pHMEM) &num_prhidden);

      mem_alloc(size, &prvisible, (void *pHMEM) &num_prvisible);
   }

   for (loop = 0; loop < size; loop++)
   {
      num_visible[loop] = (BYTE)ZERO;
      num_hidden[loop] = (BYTE)ZERO;

      if (settings.prod_type == VIP)
      {
         num_prvisible[loop] = (BYTE)ZERO;
         num_prhidden[loop] = (BYTE)ZERO;
      }
   }

   /* determine the azimuth angle, if necessary */
   if (settings.profile)
   {
      /* allocate memory for profile colors array */
      mem_alloc((long) param.max_nhts_in_radial * sizeof(char),
                &handle.prfl, (void pHMEM) &profile_color);

      /* calculate the azimuth angle */
      dif_lat = (settings.target.loc.y - settings.observer.loc.y);
      dif_lon = (settings.target.loc.x - settings.observer.loc.x);

      if (dif_lat == (GFLOAT)ZERO)
      {
         if (dif_lon == (GFLOAT)ZERO)
            azimuth_start = azimuth_end = (GFLOAT)90;
         else
            azimuth_start = azimuth_end = (GFLOAT)270;
      }
      else
      {
         dif_lon_m      = dif_lon * METERS_PER_DEGREE *
                          cos (settings.observer.loc.y / DEG_PER_RAD);
         dif_lat_m      = dif_lat * METERS_PER_DEGREE;

         azimuth_start  = atan2 (dif_lon_m, dif_lat_m) * DEG_PER_RAD;

         if (azimuth_start < (GFLOAT)ZERO)
            azimuth_start += (GFLOAT)360;

         azimuth_end    = azimuth_start;

         settings.start = settings.end = azimuth_start;
      }
   }
   else
   {
      /* assign the azimuth angles */
      azimuth_start = (GFLOAT) settings.start;
      azimuth_end   = (GFLOAT) settings.end;
   }

   /* initialize the percent indicator */
   next_step = 10;

   /* set the initial distance and elevation */
   distance[ZERO] = (GFLOAT)ZERO;

   lat = settings.observer.loc.y;
   lon = settings.observer.loc.x;

   elev_hgts[ZERO]  = interp_elev (dted, lat, lon);

   /* calculate the quadrant angles */
   if (param.y_res_m > param.x_res_m)
      quad1 = (GFLOAT)90 -
              (atan(param.y_res_m/param.x_res_m) * DEG_PER_RAD);
   else
      quad1 = atan(param.y_res_m/param.x_res_m) * DEG_PER_RAD;

   quad2 = (GFLOAT)180 - quad1;
   quad3 = (GFLOAT)180 + quad1;
   quad4 = (GFLOAT)360 - quad1;

   /* build elevation array for each radial  */
   for (radial_azimuth = azimuth_start, param.radial_count = 0;
        radial_azimuth <= azimuth_end;
        radial_azimuth += param.radial_spacing, param.radial_count++)
   {
      make_one_radial (radial_azimuth, quad1, quad2, quad3, quad4);

      if (!settings.profile)
      {
         percent = (radial_azimuth - azimuth_start) /
                   (azimuth_end - azimuth_start) * (GFLOAT)100;

         if ((percent >= next_step) /*&& (los_win != (WINDOW)NULL)*/)
         {
            sprintf(message, "%2ld%%", (long)next_step);
#if 0
            xvt_vobj_set_title (xvt_win_get_ctl(los_win, LOS_DLG_TEXT_2), message);
#endif
            next_step += 10;
         }
#if 0
         xvt_app_process_pending_events();
#endif
         if (wininfo.cancel_button_status)
            break;
      }
   }
 
   if ((!settings.profile) && (!wininfo.cancel_button_status))
   {
      /* make sure to get 4 corner quadrant angles */
      for (loop = 1; loop <= 2; loop++)
      {
         if (loop == 2)
         {
            quad1 += (GFLOAT)360;
            quad2 += (GFLOAT)360;
            quad3 += (GFLOAT)360;
            quad4 += (GFLOAT)360;
         }

         if ((quad1 > azimuth_start) && (quad1 < azimuth_end))
         {
            make_one_radial (quad1, quad1, quad2, quad3, quad4);
#if 0
            xvt_app_process_pending_events();
#endif
            if (wininfo.cancel_button_status)
               break;
         }

         if ((quad2 > azimuth_start) && (quad2 < azimuth_end))
         {
            make_one_radial (quad2, quad1, quad2, quad3, quad4);
#if 0
            xvt_app_process_pending_events();
#endif
            if (wininfo.cancel_button_status)
               break;
         }

         if ((quad3 > azimuth_start) && (quad3 < azimuth_end))
         {
            make_one_radial (quad3, quad1, quad2, quad3, quad4);
#if 0
            xvt_app_process_pending_events();
#endif
            if (wininfo.cancel_button_status)
               break;
         }

         if ((quad4 > azimuth_start) && (quad4 < azimuth_end))
         {
            make_one_radial (quad4, quad1, quad2, quad3, quad4);
#if 0
            xvt_app_process_pending_events();
#endif
            if (wininfo.cancel_button_status)
               break;
         }
      }
   }

   if ((!settings.profile) && (!wininfo.cancel_button_status))
   {
      if (settings.prod_type == VIS)
      {
         for (loop = 0; loop < size; loop++)
         {
            if ((num_visible[loop] == (BYTE)ZERO) &&
                (num_hidden[loop] == (BYTE)ZERO))
               continue;

            if (prod1_ptr[loop] != (char)ZERO)
               continue;

            if (num_visible[loop] >= num_hidden[loop])
               prod1_ptr[loop] = (char)DEF_VISIBLE;
            else
               prod1_ptr[loop] = (char)DEF_HIDDEN;
         }
      }
      else
      {
         for (loop = 0; loop < size; loop++)
         {
            if ((num_visible[loop] == (BYTE)ZERO) &&
                (num_hidden[loop] == (BYTE)ZERO) &&
                (num_prvisible[loop] == (BYTE)ZERO) &&
                (num_prhidden[loop] == (BYTE)ZERO))
               continue;

            if (prod1_ptr[loop] != (char)ZERO)
               continue;

            maxcheck = num_visible[loop];
            maxcheck = (BYTE) maxof (maxcheck, num_hidden[loop]);
            maxcheck = (BYTE) maxof (maxcheck, num_prvisible[loop]);
            maxcheck = (BYTE) maxof (maxcheck, num_prhidden[loop]);

            if (maxcheck == num_visible[loop])
               prod1_ptr[loop] = (char)DEF_VISIBLE;
            else if (maxcheck == num_hidden[loop])
               prod1_ptr[loop] = (char)DEF_HIDDEN;
            else if (maxcheck == num_prvisible[loop])
               prod1_ptr[loop] = (char)PROB_VISIBLE;
            else
               prod1_ptr[loop] = (char)PROB_HIDDEN;
         }
      }
   }

   /* free the product check arrays */
   if (xvt_gmem_unlock(visible) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock visible matrix memory");
   if (xvt_gmem_unlock(hidden) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock hidden matrix memory");
   if (xvt_gmem_free (visible) == FALSE)
      xvt_dm_post_fatal_exit("Unable to free visible matrix memory");
   if (xvt_gmem_free (hidden) == FALSE)
      xvt_dm_post_fatal_exit("Unable to free hidden matrix memory");

   if (settings.prod_type == VIP)
   {
      if (xvt_gmem_unlock(prvisible) == FALSE)
         xvt_dm_post_fatal_exit("Unable to unlock prvisible matrix memory");
      if (xvt_gmem_unlock(prhidden) == FALSE)
         xvt_dm_post_fatal_exit("Unable to unlock prhidden matrix memory");
      if (xvt_gmem_free (prvisible) == FALSE)
         xvt_dm_post_fatal_exit("Unable to free prvisible matrix memory");
      if (xvt_gmem_free (prhidden) == FALSE)
         xvt_dm_post_fatal_exit("Unable to free prhidden matrix memory");
   }

   /* unlock the trig tables */
   if (xvt_gmem_unlock(trig_table.shandle) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock trig table memory");
   if (xvt_gmem_unlock(trig_table.chandle) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock trig table memory");

   /* unlock the dted memory */
   if (xvt_gmem_unlock (elev_raster.data_matrix.ghandle) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock DTED memory");

   /* unlock product memory */
   if (xvt_gmem_unlock(prod_raster->data_matrix.ghandle) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock product memory in vip_mask()");

   /* unlock and free memory */
   if (settings.profile)
      if (xvt_gmem_unlock (handle.prfl) == FALSE)
         xvt_dm_post_fatal_exit("Unable to unlock profile color memory");
   if (xvt_gmem_unlock (handle.ydim) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock elevation buffer y-dimension");
   if (xvt_gmem_free (handle.ydim) == FALSE)
      xvt_dm_post_fatal_exit("Unable to free elevation buffer y-dimension");
   if (xvt_gmem_unlock (handle.xdim) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock elevation buffer x-dimension");
   if (xvt_gmem_free (handle.xdim) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock elevation buffer x-dimension");
   if (xvt_gmem_unlock (handle.hgts) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock elevation heights buffer");
   if (!settings.profile)
      if (xvt_gmem_free (handle.hgts) == FALSE)
         xvt_dm_post_fatal_exit("Unable to free elevation heights buffer");
   if (xvt_gmem_unlock (handle.dist) == FALSE)
      xvt_dm_post_fatal_exit("Unable to unlock distance buffer");
   if (xvt_gmem_free (handle.dist) == FALSE)
      xvt_dm_post_fatal_exit("Unable to free distance buffer");

   handle.ydim = (GHANDLE)NULL;
   handle.xdim = (GHANDLE)NULL;
   handle.dist = (GHANDLE)NULL;

   if (!settings.profile)
      handle.hgts = (GHANDLE)NULL;
}

/************************************************************************
 *   FUNCTION     : make_one_radial
 *
 *   PURPOSE      : constructs a single radial in the product
 *
 *   INPUTS       : radial_azimuth - the current azimuth angle for the
 *                                   radial being built
 *                  quad1          - 1st quadrant boundary angle
 *                  quad2          - 2nd quadrant boundary angle
 *                  quad3          - 3rd quadrant boundary angle
 *                  quad4          - 4th quadrant boundary angle
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : make_radials in module MAKE_RDL.C
 *
 *   CALLS        : obsv_elev in module ELEVATIO.C
 *                  vis_mask in module VIS_MASK.C
 *                  vip_mask in module VIP_MASK.C
 *
 ************************************************************************/

#ifdef PROTO
static void make_one_radial (GFLOAT radial_azimuth, GFLOAT quad1,
                             GFLOAT quad2, GFLOAT quad3, GFLOAT quad4)
#else
static void make_one_radial (radial_azimuth, quad1, quad2, quad3, quad4)
GFLOAT radial_azimuth, quad1, quad2, quad3, quad4;
#endif
{
   GFLOAT corr_angle;                    /* angular measurement         */
   GFLOAT new_radial_length;             /* length of radial in meters  */
   GFLOAT sin_azimuth;                   /* sine of the azimuth         */
   GFLOAT cos_azimuth;                   /* cosine of the azimuth       */
   GFLOAT angle_dist;                    /* angular radial measure      */
   GFLOAT dlat2, dlon2, dlat3;           /* end lat/lon along radial    */
   GFLOAT temp;                          /* temp storage                */
   GFLOAT lat, lon;                      /* current lat/lon             */
   GFLOAT lat_inc, lon_inc;              /* lat/lon increment in radial */

   long index;                           /* elevation heights index     */
   long px, py;                          /* array indices               */

   if (radial_azimuth > (GFLOAT)360)
      radial_azimuth -= (GFLOAT)360;
   if (radial_azimuth < (GFLOAT)0)
      radial_azimuth += (GFLOAT)360;
   if (radial_azimuth == (GFLOAT)360)
      radial_azimuth = (GFLOAT)0;

   cos_azimuth = -1 * cosine[(long)(radial_azimuth * 10.0)];
   sin_azimuth = sine[(long)(radial_azimuth * 10.0)];

   corr_angle = (GFLOAT)90 - radial_azimuth;
   if (corr_angle < (GFLOAT)ZERO)
      corr_angle += (GFLOAT)360;

   /* calculate the radial step size for specific azimuth */
   if ((radial_azimuth > quad4) || (radial_azimuth < quad1) ||
      ((radial_azimuth > quad2) && (radial_azimuth < quad3)))
      param.radial_step_size = param.y_res_m / cosine[(long)(radial_azimuth * 10.0)];
   else
      param.radial_step_size = param.x_res_m / cosine[(long)(corr_angle * 10.0)];

   if (param.radial_step_size < (GFLOAT)0)
      param.radial_step_size = (GFLOAT)0 - param.radial_step_size;

   /* calculate the number of steps along the radial */
   param.nhts_in_radial = param.orig_radial_length / param.radial_step_size;

   new_radial_length = param.radial_step_size * param.nhts_in_radial;

   /* calculate the lat/lon pair of the final point along the radial */
   angle_dist = new_radial_length / METERS_PER_DEGREE;
   dlat2 = dlat3 = settings.observer.loc.y + angle_dist * cos_azimuth;

   if (dlat3 < (GFLOAT)0)
      dlat3 += (GFLOAT)360;

   temp = cosine[(long)(dlat3 * 10.0)];

   if (temp == (GFLOAT)ZERO) 
      dlon2 = settings.observer.loc.x;
   else
      dlon2 = settings.observer.loc.x + angle_dist * sin_azimuth / temp;

   if (dlon2 < (GFLOAT)-180)
      dlon2 += (GFLOAT)360;
   else if (dlon2 > (GFLOAT)180)
      dlon2 -= (GFLOAT)360;

   lat_inc = (dlat2 - settings.observer.loc.y) / param.nhts_in_radial;
   lon_inc = (dlon2 - settings.observer.loc.x) / param.nhts_in_radial;

   lat = settings.observer.loc.y;
   lon = settings.observer.loc.x;

   /* for each point along the radial */
   for (index = ONE; index <= param.nhts_in_radial; index++)
   {
      lat += lat_inc;
      lon += lon_inc;

      distance[index] = distance[index-ONE] + param.radial_step_size;

      /* create product buffer items */
      px = (long)(prod_raster->data_matrix.width / (GFLOAT)2 -
           (settings.observer.loc.x - lon) /
           prod_raster->data_hdr.x_res);
      py = (long)(prod_raster->data_matrix.height / (GFLOAT)2 -
           (lat - settings.observer.loc.y) /
           prod_raster->data_hdr.y_res);

      prod_buff_xdim[index] =
               ((px < (long)ZERO) ||
               (px >= prod_raster->data_matrix.width)) ?
               (long)-1 : px;

      prod_buff_ydim[index] =
               ((py < (long)ZERO) ||
               (py >= prod_raster->data_matrix.height)) ?
               (long)-1 : py;

      /* get elevation point along radial */
      angle_dist = distance[index] / METERS_PER_DEGREE;
      dlat2 = dlat3 = settings.observer.loc.y + angle_dist * (-1*cos_azimuth);

      if (dlat3 < (GFLOAT)0)
         dlat3 += (GFLOAT)360;

      temp = cosine[(long)(dlat3 * 10.0)];

      if (temp == (GFLOAT)ZERO)
         dlon2 = settings.observer.loc.x;
      else
         dlon2 = settings.observer.loc.x + angle_dist *
                       sin_azimuth / temp;

      if (dlon2 < (GFLOAT)-180)
         dlon2 += (GFLOAT)360;
      else if (dlon2 > (GFLOAT)180)
         dlon2 -= (GFLOAT)360;

      elev_hgts[index] = interp_elev (dted, dlat2, dlon2);
   }

   /* call appropriate product routine */
   if (settings.prod_type == VIS)
      vis_mask(elev_hgts, prod_buff_ydim, prod_buff_xdim,
              (char *pHMEM) &profile_color);
   else
      vip_mask(elev_hgts, prod_buff_ydim, prod_buff_xdim,
              (char *pHMEM) &profile_color);
}
